import re


class Action:
    def __init__(self, action_config, namespace=None):
        self.action_config = action_config
        self.namespace = namespace
        self.setup()

    def setup(self):
        matcher = re.compile(r"(?:\{\{)([a-zA-Z0-9\.]+)(?:\}\})")
        replacer = "{0.\\1}"

        for config in self.action_config:
            config_data = self.action_config[config]
            if isinstance(config_data, str):
                config_data = matcher.sub(replacer, config_data).format(self.namespace)
            setattr(self, config, config_data)

    def execute(self):

        if self.type == "log":
            print(self.action_eval(self.value))
        elif self.type == "conditional":
            if eval(self.condition):
                return self.action_eval(self.true)
            else:
                return self.action_eval(self.false)
        elif self.type == "store":
            setattr(
                self.namespace, self.action_eval(self.name), self.action_eval(self.data)
            )
        elif self.type == "multi":
            for action_config in self.value:
                self.action_eval(action_config)
        elif self.type == "math":
            return eval(self.action_eval(self.value))

    def action_eval(self, f):
        if isinstance(f, dict):
            return Action(f, namespace=self.namespace).execute()
        else:
            return f



import os
from typing import Dict, List

from discord import (
    Intents,
    Game,
    Activity,
    Streaming,
    ActivityType,
    AllowedMentions,
    Embed,
)
from discord.ext.commands import (
    Bot,
    Command,
    Context,
    when_mentioned,
    when_mentioned_or,
)

from .action import Action
from .utils import apply_func_to_all_strings, spformat

def build_bot_from_config(config: Dict) -> Bot:
    new_config = {}

    def get_prefix(bot, msg):
        prefixes = config.get("prefix") or config.get("prefixes")
        if not isinstance(prefixes, list):
            prefixes = [prefixes]
        if prefixes:
            mentioned = "{{when_mentioned}}"
            if mentioned in prefixes:
                return when_mentioned_or(*(set(prefixes) - {mentioned}))(bot, msg)
            else:
                return prefixes
        else:
            return when_mentioned(bot, msg)

    new_config["command_prefix"] = get_prefix
    new_config["owner_id"] = config.get("owner")
    new_config["description"] = config.get("description")
    new_config["case_insensitive"] = not config.get("case_sensitive", True)
    new_config["intents"] = (
        Intents.all() if config.get("intents") == "all" else Intents.default()
    )
    presence_data = config.get("presence") or config.get("activity")
    if presence_data:
        presence_type = presence_data.get("type", "playing")
        text = presence_data.get("text")

        if presence_type == "watching":
            activity = Activity(type=ActivityType.watching, name=text)
        elif presence_type == "listening":
            activity = Activity(type=ActivityType.listening, name=text)
        elif presence_type == "streaming":
            activity = Streaming(name=text, url=presence_data.get("url"))
        else:
            activity = Game(name=text)
        new_config["activity"] = activity

    allowed_input = config.get("allowed_mentions")
    if allowed_input:
        allowed = AllowedMentions.none()

        if isinstance(allowed_input, str):
            if allowed_input == "all":
                allowed = AllowedMentions.all()
        else:
            for f in allowed_input:
                setattr(allowed, f, True)
        new_config["allowed_mentions"] = allowed

    bot = Bot(**new_config)

    if "jishaku" in config.get("include", []):
        try:
            bot.load_extension("jishaku")
        except Exception:
            raise ModuleNotFoundError(
                "Unable to use jishaku. "
                "jishaku is required if you want to include it, install using\n"
                "pip install jishaku"
            )

    return bot


def add_commands(bot: Bot, command_configs: List[Dict]):
    for command_config in command_configs:
        bot.add_command(create_command(command_config))


def create_command(command_config: Dict) -> Command:
    defined_args = command_config.get("args", [])
    async def command(ctx: Context, *args):
        def fill(string):
            return spformat(string, ctx)
        local_command_config = apply_func_to_all_strings(command_config, fill)
        for i, arg in enumerate(defined_args):
            if arg.startswith("*"):
                setattr(ctx, arg[1:], " ".join(args[i:]))
                break
            setattr(ctx, arg, args[i])

        action = local_command_config.get("action")
        if action:
            Action(action, namespace=ctx).execute()

        resp = local_command_config.get("response")
        if resp:
            await ctx.send(fill(resp))
        rep = local_command_config.get("reply")
        if rep:
            await ctx.reply(fill(rep))
        embed = local_command_config.get("embed")
        if embed:
            final_embed = Embed.from_dict(embed)
            await ctx.send(embed=final_embed)

    aliases = command_config.get("aliases", [])
    if isinstance(aliases, str):
        aliases = [aliases]

    return Command(
        command,
        name=command_config["name"],
        aliases=aliases,
        description=command_config.get("description", ""),
    )


def add_listeners(bot: Bot, event_configs: List[Dict]):
    for event_config in event_configs:
        bot.add_listener(*create_event(event_config))


def create_event(event_config: Dict):
    async def event(*args, **kwargs):
        action = event_config.get("action")
        if action:
            Action(action).execute()

    return (event, event_config["name"])


class Client:
    def __init__(self, bot_data):
        if isinstance(bot_data, str):
            self.bot_data = self.load(bot_data)
        else:
            self.bot_data = bot_data
        self.config = self.bot_data.get("config")
        self.commands = self.bot_data.get("commands") or self.bot_data.get(
            "command", []
        )
        self.events = self.bot_data.get("events") or self.bot_data.get("event", [])
        self.bot = build_bot_from_config(self.config)
        add_commands(self.bot, self.commands)
        add_listeners(self.bot, self.events)

    def run(self):
        if self.config["token"] == "{{env.token}}":
            try:
                from dotenv import load_dotenv
            except ImportError:
                raise ModuleNotFoundError(
                    "Unable to import dotenv. "
                    "dotenv is required for using env files, install using\n"
                    "pip install python-dotenv"
                )
            else:
                load_dotenv(dotenv_path=os.path.join(os.getcwd(),".env"))
            self.config["token"] = os.getenv("TOKEN") or os.getenv("token")

        self.bot.run(self.config["token"])

    def load(self, fp) -> Dict:
        with open(fp, "rb") as f:
            _, filetype = os.path.splitext(fp)
            if filetype == ".json":
                import json

                result = json.load(f)
            elif filetype == ".xml":
                from .utils import parse_xml_to_dict

                result = parse_xml_to_dict(f)
            elif filetype == ".yaml":
                try:
                    import yaml
                except ModuleNotFoundError:
                    raise ModuleNotFoundError(
                        "Unable to import yaml. "
                        "xmltodict is required for using yaml files, install using\n"
                        "pip install PyYAML"
                    )
                result = yaml.safe_load(f)
            else:
                raise Exception("Unsupported filetype")

        return result



import json
import re

def change_case(string):
    return "".join(["_" + i.lower() if i.isupper() else i for i in string]).lstrip("_")


def change_case_for_dict_keys(dictionary):
    new_dict = {}
    for key in dictionary:
        if isinstance(dictionary[key], dict):
            dictionary[key] = change_case_for_dict_keys(dictionary[key])
        elif isinstance(dictionary[key], list):
            for possible_dictionary in [
                x for x in dictionary[key] if isinstance(x, dict)
            ]:
                dictionary[key].remove(possible_dictionary)
                dictionary[key].append(change_case_for_dict_keys(possible_dictionary))
        new_dict[change_case(key)] = dictionary[key]

    return new_dict


def convert_string_bools_to_bools(dictionary):
    new_dict = {}
    for key in dictionary:
        if isinstance(dictionary[key], dict):
            dictionary[key] = convert_string_bools_to_bools(dictionary[key])
        elif isinstance(dictionary[key], list):
            for value in dictionary[key]:
                if isinstance(value, dict):
                    dictionary[key].remove(value)
                    dictionary[key].append(convert_string_bools_to_bools(value))
                elif isinstance(value, str):
                    if value.lower() == "true":
                        value = True
                    elif value.lower() == "false":
                        value = False
        elif isinstance(dictionary[key], str):
            if dictionary[key].lower() == "true":
                dictionary[key] = True
            elif dictionary[key].lower() == "false":
                dictionary[key] = False
        new_dict[key] = dictionary[key]

    return new_dict


def parse_xml_to_dict(xml_bytes):
    try:
        import xmltodict
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "Unable to import xmltodict. "
            "xmltodict is required for using xml files, install using\n"
            "pip install xmltodict"
        )
    parsed_xml = xmltodict.parse(xml_bytes)
    parsed_xml = json.loads(json.dumps(parsed_xml))
    final_dict = convert_string_bools_to_bools(change_case_for_dict_keys(parsed_xml))
    return final_dict.get("root")

def apply_func_to_all_strings(dictionary_or_list, func):
    new_dict = {}
    new_list = []
    if isinstance(dictionary_or_list, list):
        for item in dictionary_or_list:
            if isinstance(item, str):
                new_list.append(func(item))
            else:
                new_list.append(apply_func_to_all_strings(item, func))
    elif isinstance(dictionary_or_list, dict):
        for item in dictionary_or_list:
            if isinstance(dictionary_or_list[item], str):
                new_dict[item] = func(dictionary_or_list[item])
            else:
                new_dict[item] = apply_func_to_all_strings(dictionary_or_list[item], func)
    
    return new_dict or new_list or dictionary_or_list

def spformat(string, namespace):
    matcher = re.compile(r"(?:\{\{)([a-zA-Z0-9\.]+)(?:\}\})")
    replacer = "{0.\\1}"
    final = matcher.sub(replacer, string).format(namespace)


    # math nodes
    matcher = re.compile(r"(?:<<)([0-9\. \+\-/\*\(\)]+)(?:>>)")

    for match in re.findall(matcher,final):
        evaled_match = eval(match)
        final = final.replace(f"<<{match}>>", str(evaled_match), 1)
    return final



import setuptools
import re 

version = ''
with open('discordlib/__init__.py') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

with open("README.md", "r", encoding="utf-8") as readme_file:
    long_description = readme_file.read()

with open("requirements.txt", "r", encoding="utf-8") as req_file:
    requirements = req_file.readlines()

extras_require = {
    'xml': ['xmltodict'],
    'yaml': ['PyYAML'],
}

setuptools.setup(
    name="discord-lib",
    version=version,
    author="Swas.py",
    license="MIT",
    description="Create discord.py clients from XML, JSON, yaml etc",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/CodeWithSwastik/discord-lib",
    packages=["discordlib"],
    install_requires=requirements,
    extras_require=extras_require,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
    python_requires=">=3.6",
)



from argparse import ArgumentParser
from pathlib import Path

from .client import Client

def main():
    parser = ArgumentParser("discord-lib")
    parser.add_argument("filename", type=Path)
    args = parser.parse_args()
    fp = args.filename
    bot = Client(str(fp))
    bot.run()

if __name__ == '__main__':
    main()
